/*

INTMPF.cc -- source file for the reference backend of the iRRAM library

Copyright (C) 2004 Norbert Mueller

This file is part of the iRRAM Library.

The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#include "iRRAM_core.h"
#include <math.h>
#include <assert.h>

INTEGER denorm (const INT_MPF  &x,const INTEGER &p){
   INTEGER i= x.exponent - p;
   if ( i <= 0 ) return 0;
   if ( i > 2147000000 ) exit(1);
   return scale(x.mantissa,(long) (1-length(x.mantissa)+i));  
};

INT_MPF norm (const INTEGER &x,const INTEGER &p){
  return INT_MPF(x,p-1+length(x));
};

int sign         (const INT_MPF &x){
  if (x.mantissa>0) return 1;
  if (x.mantissa<0) return -1;
  return 0;
};

INTEGER size (const INT_MPF &x){
  if ( sign(x) ==  0 ) exit(1);
  return x.exponent;
};


INT_MPF add  (const INT_MPF &x,
              const INT_MPF &y,
              const INTEGER &p){
INTEGER a = denorm(x,p-1);
INTEGER b = denorm(y,p-1);
INTEGER c = a+b;
return norm(c,p-1);
};

INT_MPF sub  (const INT_MPF &x,
              const INT_MPF &y,
              const INTEGER &p){
INTEGER a = denorm(x,p-1);
INTEGER b = denorm(y,p-1);
INTEGER c = a-b;
return norm(c,p-1);
};

INT_MPF mult (const INT_MPF &x,
              const INT_MPF &y, 
              const INTEGER &p){
  if (sign(x)==0 || sign (y)==0 ) return INT_MPF(INTEGER(0));
  if (p > size(x) + size(y) )     return INT_MPF(INTEGER(0));
  INTEGER ex = p-size(y)-2; 
  INTEGER ey = p-size(x)-2;
  INTEGER e  = ex+ey;
  INTEGER a  = denorm(x,ex);  
  INTEGER b  = denorm(y,ey);
  INTEGER c  = a*b;
  INT_MPF z = norm(c,e); 
  return z;
}

INT_MPF inv (const INT_MPF &x,
             const INTEGER &p){
  if (sign(x)==0)    exit(1);
  if (p > -size(x) ) return INT_MPF(INTEGER(0));
  INTEGER ex = p + 2*size(x) - 5; 
  INTEGER ez = ex -size(x) + p - 1 ;
  INTEGER a  = denorm(x,ex);  
  INTEGER c  = scale(INTEGER(1),-ez-ex)/a;
  INT_MPF z = norm(c,ez); 
  return z;
}

INT_MPF div (const INT_MPF &x,
             const INT_MPF &y,
             const INTEGER &p){
  if (sign(y)==0)    exit(1);
  if (sign(x)==0)    return INT_MPF(INTEGER(0));
  if (p > size(x)-size(y) ) return INT_MPF(INTEGER(0));
  INTEGER ey = p + 2*size(y) - 5; 
  INTEGER ez = ey -size(y) + p - 1 ;
  INTEGER a  = denorm(y,ey);  
  INTEGER c  = scale(INTEGER(1),-ez-ey)/a;
  INT_MPF z = norm(c,ez); 
  return z;
}

INT_MPF abs  (const INT_MPF &x){
  if ( sign(x)>= 0 ) return x;
  else return INT_MPF(-x.mantissa,x.exponent);
}

/*********************check!**************************/
void getsize (const INT_MPF &x, sizetype &s){
  if ( sign(x)== 0 ) {
    s.exponent=min_exponent;
    s.mantissa=0;
  } else {
    s.exponent=size(x)-GUARD_BITS;
    s.mantissa=denorm(x,s.exponent);
  }  
}

/*********************check!**************************/
INT_MPF double2INT_MPF(const double &d){
/*abort();
*/
  if ( !isfinite(d) ) throw iRRAM_Numerical_Exception(iRRAM_conversion_from_infinite);
  int    exponent;
  double d_mantissa=frexp(d,&exponent);
  INTEGER i_mantissa =ldexp(d_mantissa,60);
  return norm(i_mantissa, INTEGER(exponent-60));
};

/*********************check!**************************/
INT_MPF shift(const INT_MPF &x, const INTEGER& n){
((*((int*)NULL))++);
assert(false);
/*
return INT_MPF(x.mantissa,x.exponent+n);
*/
}

/*********************check!**************************/
void write   (const INT_MPF &x,int w){
/*assert(false);
*/
   if ( sign(x)==0 ) {
      printf("0");
      for (int i=1;i<w;i++) printf(" ");
      return;
   }
   rwrite(denorm(x,size(x)),w-8);
   printf("*2**");
   rwrite(size(x),4);
}
